"""
Provide implementation of the error for PyPi version checking.
"""
NOT_SUPPORTED_CONTINUOUS_INTEGRATION_ERROR_MESSAGE = 'Continuous integration `{ci_name}` is not supported.'


class NotSupportedContinuousIntegrationError(Exception):
    """
    Not supported continuous integration error.
    """

    def __init__(self, message):
        self.message = message



"""
Provide fixtures and useful functionality for testing.
"""
import pytest


class Response:
    """
    Implements response fixture.
    """

    def __init__(self, status_code=None):
        self._status_code = status_code

    @property
    def status_code(self):
        """
        Get response status code.
        """
        return self._status_code

    @status_code.setter
    def status_code(self, status_code_):
        """
        Set response status code.
        """
        self._status_code = status_code_


@pytest.fixture
def response():
    """
    Get response object fixture.
    """
    return Response()



"""
Provide implementation of the command line interface.
"""
import sys

import click

from pypi_version.config import ConfigFile
from pypi_version.constants import (
    PYPI_VERSION_CHECKING_FAILED_SYS_CODE,
    PYPI_VERSION_CHECKING_PASSED_SYS_CODE,
)
from pypi_version.main import (
    PullRequest,
    PypiPackageVersion,
)
from pypi_version.utils import (
    get_package_to_check_version,
    install_package_to_check,
)


@click.group()
@click.version_option(version='0.1.0')
@click.help_option()
def cli():
    """
    Command line interface for PyPi version checking.
    """
    pass


@cli.command()
def check():
    """
    Check if you haven't forgotten to bump the PyPi package version.
    """
    install_package_to_check()

    config = ConfigFile().parse()
    parsed_package_version = get_package_to_check_version(name=config.package_name)

    if not PullRequest().is_for_check(
            ci_name=config.ci_name,
            develop_branch=config.development_branch,
            release_branch=config.release_branch,
    ):
        sys.exit(PYPI_VERSION_CHECKING_PASSED_SYS_CODE)

    if PypiPackageVersion.does_exist(name=config.package_name, version=parsed_package_version):
        sys.exit(PYPI_VERSION_CHECKING_FAILED_SYS_CODE)

    sys.exit(PYPI_VERSION_CHECKING_PASSED_SYS_CODE)



"""
Provide constants for PyPi version checking.
"""
PYPI_VERSION_CHECKING_PASSED_SYS_CODE = 0
PYPI_VERSION_CHECKING_FAILED_SYS_CODE = -1

PYPI_VERSION_CONFIG_FILE_NAME = 'pypi-version'
FETCH_PYPI_PACKAGE_URL = 'https://pypi.org/project/{package_name}/{package_version}/'

HTTP_STATUS_OK = 200


class TravisCi:
    """
    Travis continuous integration class with related data.
    """

    NAME = 'travis'
    ENV_VARIABLE_PR_BRANCH_FROM_NAME = 'TRAVIS_PULL_REQUEST_BRANCH'
    ENV_VARIABLE_PR_BRANCH_TO_NAME = 'TRAVIS_BRANCH'



"""
Provide tests for implementation of the PyPi version checking.
"""
import os

import pytest

from pypi_version.constants import (
    HTTP_STATUS_OK,
    TravisCi,
)
from pypi_version.errors import (
    NOT_SUPPORTED_CONTINUOUS_INTEGRATION_ERROR_MESSAGE,
    NotSupportedContinuousIntegrationError,
)
from pypi_version.main import (
    PullRequest,
    PypiPackageVersion,
)


def test_pypi_package_version_exist(mocker, response):
    """
    Case: check if PyPi package version exists.
    Expect: true is returned.
    """
    response.status_code = HTTP_STATUS_OK

    mock_request_package = mocker.patch('requests.get')
    mock_request_package.return_value = response

    result = PypiPackageVersion.does_exist(name='Django', version='2.17.0')
    assert result is True


def test_pypi_package_version_does_not_exist(mocker, response):
    """
    Case: check if PyPi package version exists.
    Expect: false is returned.
    """
    response.status_code = None

    mock_request_package = mocker.patch('requests.get')
    mock_request_package.return_value = response

    result = PypiPackageVersion.does_exist(name='Django', version='14.1.5')
    assert result is False


def test_pull_request_is_for_check():
    """
    Case: check if pull request environment matches specified in the configuration file.
    Except: true is returned.
    """
    os.environ[TravisCi.ENV_VARIABLE_PR_BRANCH_FROM_NAME] = 'develop'
    os.environ[TravisCi.ENV_VARIABLE_PR_BRANCH_TO_NAME] = 'master'

    result = PullRequest().is_for_check(ci_name='travis', develop_branch='develop', release_branch='master')
    assert result is True


def test_pull_request_is_for_check_not_supported_ci():
    """
    Case: check if pull request environment matches specified in the configuration file if CI isn't matched.
    Except: not supported continuous integration error.
    """
    expected_error_message = NOT_SUPPORTED_CONTINUOUS_INTEGRATION_ERROR_MESSAGE.format(ci_name='not-supported-ci')

    with pytest.raises(NotSupportedContinuousIntegrationError) as error:
        PullRequest().is_for_check(ci_name='not-supported-ci', develop_branch='develop', release_branch='master')

    assert expected_error_message == error.value.message


def test_pull_request_is_for_check_no_branch_to():
    """
    Case: check if pull request environment matches specified in the configuration file if is a branch job, not PR job.
    Except: false is returned.
    """
    os.environ[TravisCi.ENV_VARIABLE_PR_BRANCH_TO_NAME] = ''

    result = PullRequest().is_for_check(ci_name='travis', develop_branch='develop', release_branch='master')
    assert result is False


def test_pull_request_is_for_check_branch_from_unmatched():
    """
    Case: check if pull request environment matches specified in the configuration file if branch from isn't matched.
    Except: false is returned.
    """
    os.environ[TravisCi.ENV_VARIABLE_PR_BRANCH_FROM_NAME] = 'develop'

    result = PullRequest().is_for_check(ci_name='travis', develop_branch='development', release_branch='master')
    assert result is False


def test_pull_request_is_for_check_branch_to_unmatched():
    """
    Case: check if pull request environment matches specified in the configuration file if branch to isn't matched.
    Except: false is returned.
    """
    os.environ[TravisCi.ENV_VARIABLE_PR_BRANCH_TO_NAME] = 'release'

    result = PullRequest().is_for_check(ci_name='travis', develop_branch='development', release_branch='master')
    assert result is False



"""
Provide implementation of the PyPi version checking configurations manipulating.
"""
import os

import yaml
from accessify import private

from pypi_version.constants import PYPI_VERSION_CONFIG_FILE_NAME


class ConfigParameters:
    """
    Configuration parameters data transfer object.
    """

    def __init__(self, package_name, ci_name, development_branch, release_branch):
        self._package_name = package_name
        self._ci_name = ci_name
        self._development_branch = development_branch
        self._release_branch = release_branch

    @property
    def package_name(self):
        """
        Get configuration file's package name.
        """
        return self._package_name

    @property
    def ci_name(self):
        """
        Get configuration file's continuous integration name.
        """
        return self._ci_name

    @property
    def development_branch(self):
        """
        Get configuration file's package development branch.
        """
        return self._development_branch

    @property
    def release_branch(self):
        """
        Get configuration file's package release branch.
        """
        return self._release_branch


class ConfigFile:
    """
    Implementation of PyPi version configuration file.
    """

    @property
    def path(self):
        """
        Get the system path where parsing configuration file was called.
        """
        return os.getcwd()

    @private
    def read(self, name=PYPI_VERSION_CONFIG_FILE_NAME):
        """
        Read configuration file.

        Return dictionary.
        """
        with open(self.path + '/.' + name + '.yml') as f:
            return yaml.safe_load(f)

    def parse(self):
        """
        Parse configuration file.
        """
        config_as_dict = self.read()

        return ConfigParameters(
            package_name=config_as_dict.get('package').get('name'),
            ci_name=config_as_dict.get('ci').get('name'),
            development_branch=config_as_dict.get('branches').get('development'),
            release_branch=config_as_dict.get('branches').get('release'),
        )



"""
Setup the package.
"""
from setuptools import find_packages, setup

DESCRIPTION = \
    'Check if you haven\'t forgotten to bump the PyPi package version number before you merge a release pull request.'

with open('README.md', 'r') as read_me:
    long_description = read_me.read()

with open('requirements.txt', 'r') as f:
    requirements = f.read().splitlines()

setup(
    version='0.1.0',
    name='pypi-version',
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/dmytrostriletskyi/pypi-version',
    license='MIT',
    author='Dmytro Striletskyi',
    author_email='dmytro.striletskyi@gmail.com',
    packages=find_packages(),
    install_requires=requirements,
    entry_points={
        'console_scripts': [
            'pypi-version = pypi_version.cli:cli',
        ]
    },
    classifiers=[
        'Operating System :: OS Independent',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',

        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
)



"""
Provide implementation of the PyPi version checking.
"""
import os

import requests
from accessify import private

from pypi_version.constants import (
    FETCH_PYPI_PACKAGE_URL,
    HTTP_STATUS_OK,
    TravisCi,
)
from pypi_version.errors import (
    NOT_SUPPORTED_CONTINUOUS_INTEGRATION_ERROR_MESSAGE,
    NotSupportedContinuousIntegrationError,
)


class PypiPackageVersion:
    """
    Implementation of PyPi package version.
    """

    @staticmethod
    def does_exist(name, version):
        """
        Check if package with specified name and version already uploaded to the PyPi.

        Send request to the PyPi. If page is presented — package exist, else — does not exist.
        """
        response = requests.get(
            FETCH_PYPI_PACKAGE_URL.format(package_name=name, package_version=version),
        )

        if response.status_code == HTTP_STATUS_OK:
            return True

        return False


class PullRequest:
    """
    Implementation of pull request.
    """

    @private
    @staticmethod
    def get_ci(name):
        """
        Get continuous integration class with related data.
        """
        if name == TravisCi.NAME:
            return TravisCi

        raise NotSupportedContinuousIntegrationError(
            NOT_SUPPORTED_CONTINUOUS_INTEGRATION_ERROR_MESSAGE.format(ci_name=name),
        )

    def is_for_check(self, ci_name, develop_branch, release_branch):
        """
        Check if current pull request match pull request data in the configuration file.

        If environment variable of pull request branch you want merge to is empty, it means
        continuous integration run the job for the just branch, not pull request.

        If current pull request branches matches pull request branches in the configuration file — return true,
        else — return false.
        """
        ci = self.get_ci(name=ci_name)

        if not os.environ.get(ci.ENV_VARIABLE_PR_BRANCH_TO_NAME):
            return False

        if release_branch != os.environ.get(ci.ENV_VARIABLE_PR_BRANCH_TO_NAME):
            return False

        if develop_branch != os.environ.get(ci.ENV_VARIABLE_PR_BRANCH_FROM_NAME):
            return False

        return True



"""
Provide implementation of the utils for PyPi version checking.
"""
import os

import pkg_resources


def install_package_to_check():
    """
    Build and install the package.

    Is used to fetch the package version provided to pull request.
    """
    os.system('python3 setup.py sdist > /dev/null')
    os.system('pip3 install dist/*.tar.gz > /dev/null')


def get_package_to_check_version(name):
    """
    Get package version.
    """
    return pkg_resources.require(name)[0].version
